// whereami.cpp
// by Glenn G. Chappell
// November 2003
//
// For CS 381
// Flying interface with camera position determination.
// See function whereami for the position determination code.

#include <iostream>
using std::cerr;
using std::endl;
#include <string>
using std::string;
#include <sstream>
using std::ostringstream;
#include <iomanip>
using std::fixed;
using std::setprecision;
#include <math.h>  // Some versions of MS-Vis C++ have broken <cmath>
//using std::sqrt;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff - includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 400; // Starting window width & height, in pixels
const double neardist = 0.01; // Near & far clipping distances
const double fardist = 50.;

double winw, winh;            // Window width & height, set by reshape

// Keyboard
const int ESCKEY = 27;        // ASCII value of escape character

// Lighting
GLfloat light0_maincolor[] = { 1., 1., 1., 1. };
GLfloat light0_ambcolor[] = { 0., 0., 0., 0. };
GLfloat light0_position[] = { -2., 2., 3., 1. };
   // Above are parameters for GL_LIGHT0

// Flying interface
GLdouble viewmatrixd[16];     // Viewing matrix
bool leftdown = false;        // Is left mouse down?
bool rightdown = false;       // Is right mouse down?
int mousex, mousey;           // Saved mouse pos
const double speed = 0.006;   // Forward speed
const double rotamt = 0.001;  // Turn amt (deg/frame)
const double rollamt = 1.;    // Roll amt (deg/frame)


// whereami
// Determine camera position.
// Should be called with model/view holding proper
// viewing transformation. Returns camera position in
// reference parameters cx, cy, cz.
void whereami(double & cx, double & cy, double & cz)
{
   // Get the viewing matrix
   GLdouble modelview[16];
   glGetDoublev(GL_MODELVIEW_MATRIX, modelview);

   // Make a dummy "projection matrix"
   GLdouble projection_dummy[16];
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();
      glLoadIdentity();
      gluOrtho2D(-1., 1., -1., 1.);  // Old faithful
      glGetDoublev(GL_PROJECTION_MATRIX, projection_dummy);
   glPopMatrix();
   glMatrixMode(GL_MODELVIEW);

   // Set up a fake "viewport"
   // x & y got from -1 to 1. Thus, (0,0) is in center.
   GLint viewport_dummy[4] = { -1, -1, 2, 2 };

   GLdouble outx, outy, outz;  // Var's to save the answer in

   gluUnProject(0., 0., 0.5,
               modelview, projection_dummy, viewport_dummy,
               &outx, &outy, &outz);

   // Return the result.
   cx = outx;
   cy = outy;
   cz = outz;
}


// perprot
// Rotates about an axis in the x,y-plane, perpendicular to the given
// vector (vx, vy). Rotation amount is proportional to the length of
// the vector and to rotmultiplier.
void perprot(double rotmultiplier, double vx, double vy)
{
   double len = sqrt(vx*vx + vy*vy);

   glPushMatrix();
      glRotated(rotmultiplier*len, -vy,vx,0.);
      glMultMatrixd(viewmatrixd);
      glGetDoublev(GL_MODELVIEW_MATRIX, viewmatrixd);
   glPopMatrix();
   glutPostRedisplay();
}


// printbitmap
// Prints the given string at the given raster position
//  using GLUT bitmap fonts.
// You probably don't want any rotations in the model/view
//  transformation when calling this function.
void printbitmap(const string msg, double x, double y)
{
   glRasterPos2d(x, y);
   for (string::const_iterator ii = msg.begin();
        ii != msg.end();
        ++ii)
   {
      glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *ii);
   }
}


// setmaterial_main
// Sets up "main" material
void setmaterial_main()
{
   GLfloat paintcolor[] = { 0.6, 0.8, 0.2, 1.0 };
   GLfloat specularcolor[] = { 1.0, 1.0, 1.0, 1.0 };
   glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, paintcolor);
   glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specularcolor);
   glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.);
      // Set the shininess to 1 to find out why we like raising the
      // cosine to a high power.
}


// drawscene
// Draws the scene using the current transformations
void drawscene()
{
   const int sphereslices = 80;
   const int spherestacks = 40;

   glPushMatrix();
      glTranslated(0.8, 0., 0.);
      glutSolidSphere(0.1, sphereslices, spherestacks);
   glPopMatrix();

   glPushMatrix();
      glTranslated(-0.8, 0., 0.);
      glutSolidSphere(0.1, sphereslices, spherestacks);
   glPopMatrix();

   glPushMatrix();
      glTranslated(0., 0.8, 0.);
      glutSolidSphere(0.1, sphereslices, spherestacks);
   glPopMatrix();

   glPushMatrix();
      glTranslated(0., -0.8, 0.);
      glutSolidSphere(0.1, sphereslices, spherestacks);
   glPopMatrix();

   glPushMatrix();
      glTranslated(0., 0., -0.8);
      glutSolidSphere(0.1, sphereslices, spherestacks);
   glPopMatrix();

   glPushMatrix();
      glTranslated(0., 0., 0.8);
      glutSolidSphere(0.1, sphereslices, spherestacks);
   glPopMatrix();
}


// display
// The GLUT display function
void display()
{
   double camx, camy, camz;  // Camera position

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   // Draw scene
   glPushMatrix();
      glMultMatrixd(viewmatrixd);
      whereami(camx, camy, camz);  // Get camera location
      glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
      setmaterial_main();
      drawscene();
   glPopMatrix();

   // Text documentation
   glDisable(GL_LIGHTING);      // No lighting for text
   glDisable(GL_DEPTH_TEST);    // No depth test for text
   glMatrixMode(GL_PROJECTION);
   glPushMatrix();              // Save old projection matrix
      glLoadIdentity();
      gluOrtho2D(-1., 1., -1., 1.);  // old faithful
      glColor3d(0., 0., 0.);
      printbitmap("\"Where Am I?\"", -0.9, 0.9);
      ostringstream os;
      os << fixed << setprecision(3)
         << "Camera position: "
         << "(" << camx << ", " << camy << ", " << camz << ")";
      printbitmap(os.str(), -0.9, 0.8);
      printbitmap("A S   Roll", -0.9, 0.7);
      printbitmap("Z     Reset position", -0.9, 0.6);
      printbitmap("Esc   Quit", -0.9, 0.5);
   glPopMatrix();               // Restore old projection matrix
   glMatrixMode(GL_MODELVIEW);  // Always go back to model/view mode
   glEnable(GL_DEPTH_TEST);     // Back to "normal" for 3-D drawing
   glEnable(GL_LIGHTING);

   glutSwapBuffers();
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
   glViewport(0, 0, w, h);
   winw = w;
   winh = h;

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(60, double(w)/h, neardist, fardist);
   // When drawing text, we set up a temporary projection with
   //  gluOrtho2D, so we don't need to worry about exactly where
   //  left, right, bottom, and top are.

   glMatrixMode(GL_MODELVIEW);  // Always go back to modelview mode
}


// idle
// The GLUT idle function
void idle()
{
   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }

   // Fly
   if (leftdown || rightdown)
   {
      glPushMatrix();
         glTranslated(0., 0., rightdown ? -speed : speed);
         glMultMatrixd(viewmatrixd);
         glGetDoublev(GL_MODELVIEW_MATRIX, viewmatrixd);
      glPopMatrix();
   }
   perprot(rotamt, mousex - double(winw)/2, mousey - double(winh)/2);
   glutPostRedisplay();
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   case 'a':     // A: Left roll
   case 'A':
      glPushMatrix();
         glRotated(-rollamt, 0,0,1);
         glMultMatrixd(viewmatrixd);
         glGetDoublev(GL_MODELVIEW_MATRIX, viewmatrixd);
      glPopMatrix();
      glutPostRedisplay();
      break;
   case 's':     // S: Right roll
   case 'S':
      glPushMatrix();
         glRotated(rollamt, 0,0,1);
         glMultMatrixd(viewmatrixd);
         glGetDoublev(GL_MODELVIEW_MATRIX, viewmatrixd);
      glPopMatrix();
      glutPostRedisplay();
      break;
   case 'z':     // Z: Reset position
   case 'Z':
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
         glTranslated(0., 0., -3.);
         glGetDoublev(GL_MODELVIEW_MATRIX, viewmatrixd);
      glPopMatrix(); 
      glutPostRedisplay();
      break;
   }
}


// mouse
// The GLUT mouse function
void mouse(int button, int state, int x, int y)
{
   // Save mouse position and button states for flying code in idle
   if (button == GLUT_LEFT_BUTTON)
      leftdown = (state == GLUT_DOWN);
   if (button == GLUT_RIGHT_BUTTON)
      rightdown = (state == GLUT_DOWN);
   mousex = x;
   mousey = winh - y;
}


// motion
// The GLUT motion function
void motion(int x, int y)
{
   // Save mouse position for flying code in idle
   mousex = x;
   mousey = winh - y;
}


// passivemotion
// The GLUT passive motion function
void passivemotion(int x, int y)
{
   // Save mouse position for flying code in idle
   mousex = x;
   mousey = winh - y;
}


// init
// Initialization
// Called by main after window creation
void init()
{
   // General
   glClearColor(0.7, 0.8, 0.8, 0.0);
   glEnable(GL_DEPTH_TEST);           // 3-D so we need HSR

   // Lighting
   glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_maincolor);
   glLightfv(GL_LIGHT0, GL_SPECULAR, light0_maincolor);
   glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambcolor);
      // Light 0 position is set in display
   glEnable(GL_LIGHT0);
   glEnable(GL_LIGHTING);
   //glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
      // Un-comment the above line to try local-viewer mode

   // Viewing Matrix
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
      glTranslated(0., 0., -3.);
      glGetDoublev(GL_MODELVIEW_MATRIX, viewmatrixd);
   glPopMatrix(); 
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 381 - Where Am I?");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display);
   glutReshapeFunc(reshape);
   glutIdleFunc(idle);
   glutKeyboardFunc(keyboard);
   glutMouseFunc(mouse);
   glutMotionFunc(motion);
   glutPassiveMotionFunc(passivemotion);

   // Do something
   glutMainLoop();

   return 0;
}
